@@ -13,7 +13,13 @@ try:
 except ImportError:
     import pickle
 import random
-import sha
+
+try:
+    # Python 2.5+
+    from hashlib import sha1 as sha
+except ImportError:
+    from sha import new as sha
+
 import time
 import threading
 import types
@@ -23,16 +29,6 @@ import cherrypy
 from cherrypy.lib import http
 
 
-class PerpetualTimer(threading._Timer):
-    
-    def run(self):
-        while True:
-            self.finished.wait(self.interval)
-            if self.finished.isSet():
-                return
-            self.function(*self.args, **self.kwargs)
-
-
 missing = object()
 
 class Session(object):
@@ -40,8 +36,18 @@ class Session(object):
     
     __metaclass__ = cherrypy._AttributeDocstrings
     
-    id = None
+    _id = None
+    id_observers = None
+    id_observers__doc = "A list of callbacks to which to pass new id's."
+    
     id__doc = "The current session ID."
+    def _get_id(self):
+        return self._id
+    def _set_id(self, value):
+        self._id = value
+        for o in self.id_observers:
+            o(value)
+    id = property(_get_id, _set_id, doc=id__doc)
     
     timeout = 60
     timeout__doc = "Number of minutes after which to delete session data."
@@ -57,31 +63,46 @@ class Session(object):
     automatically on the first attempt to access session data."""
     
     clean_thread = None
-    clean_thread__doc = "Class-level PerpetualTimer which calls self.clean_up."
+    clean_thread__doc = "Class-level Monitor which calls self.clean_up."
     
     clean_freq = 5
     clean_freq__doc = "The poll rate for expired session cleanup in minutes."
     
     def __init__(self, id=None, **kwargs):
+        self.id_observers = []
         self._data = {}
         
         for k, v in kwargs.iteritems():
             setattr(self, k, v)
         
-        self.id = id
+        if id is None:
+            self.regenerate()
+        else:
+            self.id = id
+            if not self._exists():
+                # Expired or malicious session. Make a new one.
+                # See http://www.cherrypy.org/ticket/709.
+                self.id = None
+                self.regenerate()
+    
+    def regenerate(self):
+        """Replace the current session (with a new id)."""
+        if self.id is not None:
+            self.delete()
+        
+        old_session_was_locked = self.locked
+        if old_session_was_locked:
+            self.release_lock()
+        
+        self.id = None
         while self.id is None:
             self.id = self.generate_id()
             # Assert that the generated id is not already stored.
-            if self._load() is not None:
+            if self._exists():
                 self.id = None
-    
-    def clean_interrupt(cls):
-        """Stop the expired-session cleaning timer."""
-        if cls.clean_thread:
-            cls.clean_thread.cancel()
-            cls.clean_thread.join()
-            cls.clean_thread = None
-    clean_interrupt = classmethod(clean_interrupt)
+        
+        if old_session_was_locked:
+            self.acquire_lock()
     
     def clean_up(self):
         """Clean up expired sessions."""
@@ -93,7 +114,7 @@ class Session(object):
         # os.urandom not available until Python 2.4. Fall back to random.random.
         def generate_id(self):
             """Return a new session id."""
-            return sha.new('%s' % random.random()).hexdigest()
+            return sha('%s' % random.random()).hexdigest()
     else:
         def generate_id(self):
             """Return a new session id."""
@@ -119,7 +140,7 @@ class Session(object):
         data = self._load()
         # data is either None or a tuple (session_data, expiration_time)
         if data is None or data[1] < datetime.datetime.now():
-            # Expired session: flush session data (but keep the same id)
+            # Expired session: flush session data
             self._data = {}
         else:
             self._data = data[0]
@@ -128,12 +149,12 @@ class Session(object):
         # Stick the clean_thread in the class, not the instance.
         # The instances are created and destroyed per-request.
         cls = self.__class__
-        if not cls.clean_thread:
-            cherrypy.engine.on_stop_engine_list.append(cls.clean_interrupt)
+        if self.clean_freq and not cls.clean_thread:
             # clean_up is in instancemethod and not a classmethod,
-            # so tool config can be accessed inside the method.
-            t = PerpetualTimer(self.clean_freq * 60, self.clean_up)
-            t.setName("CP Session Cleanup")
+            # so that tool config can be accessed inside the method.
+            t = cherrypy.process.plugins.Monitor(
+                cherrypy.engine, self.clean_up, self.clean_freq * 60)
+            t.subscribe()
             cls.clean_thread = t
             t.start()
     
@@ -154,6 +175,10 @@ class Session(object):
         del self._data[key]
     
     def pop(self, key, default=missing):
+        """Remove the specified key and return the corresponding value.
+        If key is not found, default is returned if given,
+        otherwise KeyError is raised.
+        """
         if not self.loaded: self.load()
         if default is missing:
             return self._data.pop(key)
@@ -165,34 +190,42 @@ class Session(object):
         return key in self._data
     
     def has_key(self, key):
+        """D.has_key(k) -> True if D has a key k, else False."""
         if not self.loaded: self.load()
         return self._data.has_key(key)
     
     def get(self, key, default=None):
+        """D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None."""
         if not self.loaded: self.load()
         return self._data.get(key, default)
     
     def update(self, d):
+        """D.update(E) -> None.  Update D from E: for k in E: D[k] = E[k]."""
         if not self.loaded: self.load()
         self._data.update(d)
     
     def setdefault(self, key, default=None):
+        """D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D."""
         if not self.loaded: self.load()
         return self._data.setdefault(key, default)
     
     def clear(self):
+        """D.clear() -> None.  Remove all items from D."""
         if not self.loaded: self.load()
         self._data.clear()
     
     def keys(self):
+        """D.keys() -> list of D's keys."""
         if not self.loaded: self.load()
         return self._data.keys()
     
     def items(self):
+        """D.items() -> list of D's (key, value) pairs, as 2-tuples."""
         if not self.loaded: self.load()
         return self._data.items()
     
     def values(self):
+        """D.values() -> list of D's values."""
         if not self.loaded: self.load()
         return self._data.values()
 
@@ -217,6 +250,9 @@ class RamSession(Session):
                 except KeyError:
                     pass
     
+    def _exists(self):
+        return self.id in self.cache
+    
     def _load(self):
         return self.cache.get(self.id)
     
@@ -227,16 +263,22 @@ class RamSession(Session):
         del self.cache[self.id]
     
     def acquire_lock(self):
+        """Acquire an exclusive lock on the currently-loaded session data."""
         self.locked = True
         self.locks.setdefault(self.id, threading.RLock()).acquire()
     
     def release_lock(self):
+        """Release the lock on the currently-loaded session data."""
         self.locks[self.id].release()
         self.locked = False
+    
+    def __len__(self):
+        """Return the number of active sessions."""
+        return len(self.cache)
 
 
 class FileSession(Session):
-    """ Implementation of the File backend for sessions
+    """Implementation of the File backend for sessions
     
     storage_path: the folder where session data will be saved. Each session
         will be saved as pickle.dump(data, expiration_time) in its own file;
@@ -246,25 +288,45 @@ class FileSession(Session):
     SESSION_PREFIX = 'session-'
     LOCK_SUFFIX = '.lock'
     
-    def setup(self):
+    def __init__(self, id=None, **kwargs):
+        # The 'storage_path' arg is required for file-based sessions.
+        kwargs['storage_path'] = os.path.abspath(kwargs['storage_path'])
+        Session.__init__(self, id=id, **kwargs)
+    
+    def setup(cls, **kwargs):
+        """Set up the storage system for file-based sessions.
+        
+        This should only be called once per process; this will be done
+        automatically when using sessions.init (as the built-in Tool does).
+        """
+        # The 'storage_path' arg is required for file-based sessions.
+        kwargs['storage_path'] = os.path.abspath(kwargs['storage_path'])
+        
+        for k, v in kwargs.iteritems():
+            setattr(cls, k, v)
+        
         # Warn if any lock files exist at startup.
-        lockfiles = [fname for fname in os.listdir(self.storage_path)
-                     if (fname.startswith(self.SESSION_PREFIX)
-                         and fname.endswith(self.LOCK_SUFFIX))]
+        lockfiles = [fname for fname in os.listdir(cls.storage_path)
+                     if (fname.startswith(cls.SESSION_PREFIX)
+                         and fname.endswith(cls.LOCK_SUFFIX))]
         if lockfiles:
             plural = ('', 's')[len(lockfiles) > 1]
             warn("%s session lockfile%s found at startup. If you are "
                  "only running one process, then you may need to "
                  "manually delete the lockfiles found at %r."
-                 % (len(lockfiles), plural,
-                    os.path.abspath(self.storage_path)))
+                 % (len(lockfiles), plural, cls.storage_path))
+    setup = classmethod(setup)
     
     def _get_file_path(self):
         f = os.path.join(self.storage_path, self.SESSION_PREFIX + self.id)
-        if not os.path.normpath(f).startswith(self.storage_path):
+        if not os.path.abspath(f).startswith(self.storage_path):
             raise cherrypy.HTTPError(400, "Invalid session id in cookie.")
         return f
     
+    def _exists(self):
+        path = self._get_file_path()
+        return os.path.exists(path)
+    
     def _load(self, path=None):
         if path is None:
             path = self._get_file_path()
@@ -291,6 +353,7 @@ class FileSession(Session):
             pass
     
     def acquire_lock(self, path=None):
+        """Acquire an exclusive lock on the currently-loaded session data."""
         if path is None:
             path = self._get_file_path()
         path += self.LOCK_SUFFIX
@@ -305,6 +368,7 @@ class FileSession(Session):
         self.locked = True
     
     def release_lock(self, path=None):
+        """Release the lock on the currently-loaded session data."""
         if path is None:
             path = self._get_file_path()
         os.unlink(path + self.LOCK_SUFFIX)
@@ -331,6 +395,12 @@ class FileSession(Session):
                             os.unlink(path)
                 finally:
                     self.release_lock(path)
+    
+    def __len__(self):
+        """Return the number of active sessions."""
+        return len([fname for fname in os.listdir(self.storage_path)
+                    if (fname.startswith(self.SESSION_PREFIX)
+                        and not fname.endswith(self.LOCK_SUFFIX))])
 
 
 class PostgresqlSession(Session):
@@ -348,14 +418,32 @@ class PostgresqlSession(Session):
     
     def __init__(self, id=None, **kwargs):
         Session.__init__(self, id, **kwargs)
-        self.db = self.get_db()
         self.cursor = self.db.cursor()
     
+    def setup(cls, **kwargs):
+        """Set up the storage system for Postgres-based sessions.
+        
+        This should only be called once per process; this will be done
+        automatically when using sessions.init (as the built-in Tool does).
+        """
+        for k, v in kwargs.iteritems():
+            setattr(cls, k, v)
+        
+        self.db = self.get_db()
+    setup = classmethod(setup)
+    
     def __del__(self):
         if self.cursor:
             self.cursor.close()
         self.db.commit()
     
+    def _exists(self):
+        # Select session data from table
+        self.cursor.execute('select data, expiration_time from session '
+                            'where id=%s', (self.id,))
+        rows = self.cursor.fetchall()
+        return bool(rows)
+    
     def _load(self):
         # Select session data from table
         self.cursor.execute('select data, expiration_time from session '
@@ -378,12 +466,14 @@ class PostgresqlSession(Session):
         self.cursor.execute('delete from session where id=%s', (self.id,))
    
     def acquire_lock(self):
+        """Acquire an exclusive lock on the currently-loaded session data."""
         # We use the "for update" clause to lock the row
         self.locked = True
         self.cursor.execute('select id from session where id=%s for update',
                             (self.id,))
     
     def release_lock(self):
+        """Release the lock on the currently-loaded session data."""
         # We just close the cursor and that will remove the lock
         #   introduced by the "for update" clause
         self.cursor.close()
@@ -395,12 +485,78 @@ class PostgresqlSession(Session):
                             (datetime.datetime.now(),))
 
 
+class MemcachedSession(Session):
+    
+    # The most popular memcached client for Python isn't thread-safe.
+    # Wrap all .get and .set operations in a single lock.
+    mc_lock = threading.RLock()
+    
+    # This is a seperate set of locks per session id.
+    locks = {}
+    
+    servers = ['127.0.0.1:11211']
+    
+    def setup(cls, **kwargs):
+        """Set up the storage system for memcached-based sessions.
+        
+        This should only be called once per process; this will be done
+        automatically when using sessions.init (as the built-in Tool does).
+        """
+        for k, v in kwargs.iteritems():
+            setattr(cls, k, v)
+        
+        import memcache
+        cls.cache = memcache.Client(cls.servers)
+    setup = classmethod(setup)
+    
+    def _exists(self):
+        self.mc_lock.acquire()
+        try:
+            return bool(self.cache.get(self.id))
+        finally:
+            self.mc_lock.release()
+    
+    def _load(self):
+        self.mc_lock.acquire()
+        try:
+            return self.cache.get(self.id)
+        finally:
+            self.mc_lock.release()
+    
+    def _save(self, expiration_time):
+        # Send the expiration time as "Unix time" (seconds since 1/1/1970)
+        td = int(time.mktime(expiration_time.timetuple()))
+        self.mc_lock.acquire()
+        try:
+            if not self.cache.set(self.id, (self._data, expiration_time), td):
+                raise AssertionError("Session data for id %r not set." % self.id)
+        finally:
+            self.mc_lock.release()
+    
+    def _delete(self):
+        self.cache.delete(self.id)
+    
+    def acquire_lock(self):
+        """Acquire an exclusive lock on the currently-loaded session data."""
+        self.locked = True
+        self.locks.setdefault(self.id, threading.RLock()).acquire()
+    
+    def release_lock(self):
+        """Release the lock on the currently-loaded session data."""
+        self.locks[self.id].release()
+        self.locked = False
+    
+    def __len__(self):
+        """Return the number of active sessions."""
+        raise NotImplementedError
+
+
 # Hook functions (for CherryPy tools)
 
 def save():
     """Save any changed session data."""
     
-    if not hasattr(cherrypy._serving, "session"):
+    if not hasattr(cherrypy.serving, "session"):
         return
     
     # Guard against running twice
@@ -422,8 +578,8 @@ save.failsafe = True
 
 def close():
     """Close the session object for this request."""
-    sess = getattr(cherrypy._serving, "session", None)
-    if sess and sess.locked:
+    sess = getattr(cherrypy.serving, "session", None)
+    if getattr(sess, "locked", False):
         # If the session is still locked we release the lock
         sess.release_lock()
 close.failsafe = True
@@ -441,7 +597,8 @@ def init(storage_type='ram', path=None, path_header=None, name='session_id',
     path_header: if 'path' is None (the default), then the response
         cookie 'path' will be pulled from request.headers[path_header].
     name: the name of the cookie.
-    timeout: the expiration timeout for the cookie.
+    timeout: the expiration timeout (in minutes) for both the cookie and
+        stored session data.
     domain: the cookie domain.
     secure: if False (the default) the cookie 'secure' value will not
         be set. If True, the cookie 'secure' value will be set (to 1).
@@ -464,23 +621,50 @@ def init(storage_type='ram', path=None, path_header=None, name='session_id',
     if name in request.cookie:
         id = request.cookie[name].value
     
-    # Create and attach a new Session instance to cherrypy._serving.
+    # Find the storage class and call setup (first time only).
+    storage_class = storage_type.title() + 'Session'
+    storage_class = globals()[storage_class]
+    if not hasattr(cherrypy, "session"):
+        if hasattr(storage_class, "setup"):
+            storage_class.setup(**kwargs)
+    
+    # Create and attach a new Session instance to cherrypy.serving.
     # It will possess a reference to (and lock, and lazily load)
     # the requested session data.
-    storage_class = storage_type.title() + 'Session'
     kwargs['timeout'] = timeout
     kwargs['clean_freq'] = clean_freq
-    cherrypy._serving.session = sess = globals()[storage_class](id, **kwargs)
+    cherrypy.serving.session = sess = storage_class(id, **kwargs)
+    def update_cookie(id):
+        """Update the cookie every time the session id changes."""
+        cherrypy.response.cookie[name] = id
+    sess.id_observers.append(update_cookie)
     
+    # Create cherrypy.session which will proxy to cherrypy.serving.session
     if not hasattr(cherrypy, "session"):
         cherrypy.session = cherrypy._ThreadLocalProxy('session')
-        if hasattr(sess, "setup"):
-            sess.setup()
     
+    set_response_cookie(path=path, path_header=path_header, name=name,
+                        timeout=timeout, domain=domain, secure=secure)
+
+
+def set_response_cookie(path=None, path_header=None, name='session_id',
+                        timeout=60, domain=None, secure=False):
+    """Set a response cookie for the client.
+    
+    path: the 'path' value to stick in the response cookie metadata.
+    path_header: if 'path' is None (the default), then the response
+        cookie 'path' will be pulled from request.headers[path_header].
+    name: the name of the cookie.
+    timeout: the expiration timeout for the cookie.
+    domain: the cookie domain.
+    secure: if False (the default) the cookie 'secure' value will not
+        be set. If True, the cookie 'secure' value will be set (to 1).
+    """
     # Set response cookie
     cookie = cherrypy.response.cookie
-    cookie[name] = sess.id
-    cookie[name]['path'] = path or request.headers.get(path_header) or '/'
+    cookie[name] = cherrypy.serving.session.id
+    cookie[name]['path'] = (path or cherrypy.request.headers.get(path_header)
+                            or '/')
     
     # We'd like to use the "max-age" param as indicated in
     # http://www.faqs.org/rfcs/rfc2109.html but IE doesn't
@@ -494,6 +678,7 @@ def init(storage_type='ram', path=None, path_header=None, name='session_id',
     if secure:
         cookie[name]['secure'] = 1
 
+
 def expire():
     """Expire the current session cookie."""
     name = cherrypy.request.config.get('tools.sessions.name', 'session_id')
@@ -502,3 +687,4 @@ def expire():
     t = time.strftime("%a, %d-%b-%Y %H:%M:%S GMT", exp)
     cherrypy.response.cookie[name]['expires'] = t
 
+
